#include <math.h>
#include "mex.h"
#include "fitness_f.h"
#include "union_bs_f.h"

double fitness_f(USHORT nAtoms, USHORT nGroups, USHORT nClients,
				 mxArray *pind[], USHORT *psub, mxArray *psubclients[],
				 ULONG *pclientbdata, double *pweights)
{
	double dblFitness = 0;
	int i, j, nWords = nAtoms/ULONG_BITS;
	UBYTE *pHasProcClient = (UBYTE*)mxCalloc(nClients, sizeof(UBYTE));
	USHORT *pbad_atoms = (USHORT*)mxCalloc(nAtoms, sizeof(USHORT));

//	mexPrintf("\natoms = %u groups = %u clients = %u", nAtoms, nGroups, nClients);

	while (nGroups-- > 0)
	{
		if (*pind) // Is group not empty ?
		{
			// Compute group subscription binary data.
			mxArray *paGroupSubs = *pind;
			UINT nSubs = mxGetNumberOfElements(paGroupSubs);
			USHORT *pGSIds = (USHORT*)mxGetData(paGroupSubs);
			USHORT *pGSData = (USHORT*)mxCalloc(2*nSubs, sizeof(USHORT));
			ULONG *pGSBData = (ULONG*)mxCalloc(nWords, sizeof(ULONG));
			USHORT *pItrGSIds = pGSIds, *pItrGSData = pGSData;


//			mexPrintf("\nG%u has %u subscriptions", nGroups, nSubs);

			for (i = 0; i < nSubs; ++i)
			{
				USHORT isub = (*pItrGSIds++ - 1)<<1;
				*pItrGSData++ = psub[isub];
				*pItrGSData++ = psub[isub+1];

//				mexPrintf("\nisub = %u (%u,%u)", psub[isub], psub[isub+1]);
			}

			union_bs_f(pGSData, nSubs, pGSBData);
			mxFree(pGSData);

//			goto label1;

			// Compute bad atoms for each client in the group.
			pItrGSIds = pGSIds;

			for (i = 0; i < nSubs; ++i)
			{
				USHORT isub = *pItrGSIds++ - 1;
				USHORT nSubClients;
				USHORT *pItrCIds;

//				mexPrintf("\nisub = %u %p", isub, psubclients[isub]);

				nSubClients = mxGetNumberOfElements(psubclients[isub]);
				pItrCIds = (USHORT*)mxGetData(psubclients[isub]);

				
				while (nSubClients--)
				{
					USHORT icli = *pItrCIds++ - 1;

					if (!pHasProcClient[icli])
					{
						ULONG nword, *pItr = pGSBData, *pCBData = pclientbdata + icli * nWords;
						USHORT *pItrBadAtoms;
						pHasProcClient[icli] = 1;

//						mexPrintf("\nicli = %u", icli);

						for (j = 0; j < nWords; ++j)
						{
							nword = *pItr++ & *pCBData++;
							pItrBadAtoms = pbad_atoms + j * ULONG_BITS;

							while (nword)
							{
								if(1&nword)
									++*pItrBadAtoms;

								++pItrBadAtoms;
								nword >>= 1;
							}
						}
					}
				}
			}

			for (i = 0; i < nClients; ++i) // Clear processed client flags.
				pHasProcClient[i] = 0;

label1:
			mxFree(pGSBData);
//			goto exit;
		}

		++pind;
	}

	// Compute total weight
	for (i = 0; i < nAtoms; ++i)
	{
//		mexPrintf("\natom %u = %u", i+1, pbad_atoms[i]);
		dblFitness += pweights[i] * pbad_atoms[i];
	}

exit:
	mxFree(pbad_atoms);
	mxFree(pHasProcClient);
	return dblFitness;
}
