% (C) Mario Guimaraes 15 Dez 2001

function [ind0, cost0, indf, costf, solution, cost, iter, agendata] = genetic(pop, ...
    achosensubs, asubscriptions, aclients, aclient_bin_subs, aweights, nSubsMulticast)

global npopsize ngroups nMaxIteration useElitism pCrossover pMutation pSwitchCrossover pSwitchSelection pTournament

nMaxCrossovers = fix(npopsize ./ 2);
bRandomCrossover = 1; % Com random crossover a populaao e' mais diversificada.
bWheel = 0; % Com roullet wheel a convergencia e' maior.

% Proccess entry population
afitness_p = fitness(pop, asubscriptions, aclients, aclient_bin_subs, aweights);
indf = pop{1};
costf = afitness_p(1);

[afitness_p, ndx_p] = sort(afitness_p);
npop(:,1) = pop(ndx_p);
clear pop ndx_p;
pop = npop;
clear npop;

ind0 = pop{1};
cost0 = afitness_p(1);

% Start search
solution = ind0;
cost = cost0;
nIteration = 0;
iter = 0;
agendata = zeros(nMaxIteration, 4);
stime = clock;

while (1)
    if nIteration == fix(nMaxIteration * pSwitchCrossover), bRandomCrossover = 0; end;
    if nIteration == fix(nMaxIteration * pSwitchSelection), bWheel = 1; end;

    % Acceptance
    if nIteration == nMaxIteration | cost == 0, break; end

    % Selection
    if bWheel
        acouples = wheel(afitness_p, nMaxCrossovers);
    else
        acouples = tournament(afitness_p, nMaxCrossovers, pTournament);
    end
    
    % Two-Point Crossover
    ncand = 1;
    ahelper = uint8(zeros(1, npopsize)); % Used to avoid duplicate parents in candidate population.
    ahelper2 = uint8(zeros(1, 2 .* nMaxCrossovers)); % Registers parent individuals in candidate population.
                                                     % Used to avoid duplicate individuals in next generation.
    for pair = acouples
        if rand < pCrossover
            % Choose crosspoints p1 < p2
            p1 = achosensubs(fix(rand .* nSubsMulticast) + 1);
            p2 = achosensubs(fix(rand .* nSubsMulticast) + 1);
            if p1 > p2, tmp = p1; p1 = p2; p2 = tmp; end            

            % Crossover and generate candidate population
            if bRandomCrossover
                ind1_g = randperm(ngroups);
                ind2_g = randperm(ngroups);
            end

            for g = 1:ngroups
                if bRandomCrossover
                    g1 = pop{pair(1)}{ind1_g(g)};
                    g2 = pop{pair(2)}{ind2_g(g)};
                else
                    g1 = pop{pair(1)}{g};
                    g2 = pop{pair(2)}{g};
                end

                idx_g1 = find(p1 <= g1 & g1 <= p2);
                idx_g2 = find(p1 <= g2 & g2 <= p2);
                
                if isempty(idx_g1)
                    cpop{ncand,1}{g,1} = sort([g1 g2(idx_g2)]);
                else
                    cpop{ncand,1}{g,1} = [g1(1:idx_g1(1)-1) g2(idx_g2) g1(idx_g1(end)+1:end)];
                end
                
                if isempty(idx_g2)
                    cpop{ncand+1,1}{g,1} = sort([g2 g1(idx_g1)]);
                else                
                    cpop{ncand+1,1}{g,1} = [g2(1:idx_g2(1)-1) g1(idx_g1) g2(idx_g2(end)+1:end)];
                end
            end

            ncand = ncand + 2;
        else
            for i = pair'
                if ~ahelper(i)
                    ahelper(i) = 1;
                    cpop(ncand,1) = pop(i);
                    ahelper2(ncand) = i;
                    ncand = ncand + 1;
                end
            end
        end
    end

    % Mutation
    for i = 1:ncand-1
        if rand < pMutation
            g1 = fix(rand .* ngroups) + 1;
            
            while (1)
                g2 = fix(rand .* ngroups) + 1;
                if g1 ~= g2, break; end
            end
            
            if isempty(cpop{i}{g1})
                if ~isempty(cpop{i}{g2})
                    g2_p1 = fix(rand .* length(cpop{i}{g2})) + 1;
                    g2_p2 = fix(rand .* length(cpop{i}{g2})) + 1;            
                    if g2_p1 > g2_p2, tmp = g2_p1; g2_p1 = g2_p2; g2_p2 = tmp; end
                    
                    idxs = mutate_at(g2_p1,g2_p2);
                    cpop{i}{g1} = cpop{i}{g2}(idxs);
                    cpop{i}{g2}(idxs) = [];

                    if ahelper2(i) % Did a parent mutate ?
                        ahelper(ahelper2(i)) = 0;
                        ahelper2(i) = 0;
                    end
                end
            else
                g1_p1 = fix(rand .* length(cpop{i}{g1})) + 1;
                g1_p2 = fix(rand .* length(cpop{i}{g1})) + 1;            
                if g1_p1 > g1_p2, tmp = g1_p1; g1_p1 = g1_p2; g1_p2 = tmp; end

                idxs1 = mutate_at(g1_p1,g1_p2);

                if isempty(cpop{i}{g2})
                    cpop{i}{g2} = cpop{i}{g1}(idxs1);
                    cpop{i}{g1}(idxs1) = [];
                else
                    g2_p1 = fix(rand .* length(cpop{i}{g2})) + 1;
                    g2_p2 = fix(rand .* length(cpop{i}{g2})) + 1;            
                    if g2_p1 > g2_p2, tmp = g2_p1; g2_p1 = g2_p2; g2_p2 = tmp; end

                    idxs2 = mutate_at(g2_p1,g2_p2);
                    cpop{i}{g1} = [cpop{i}{g1} cpop{i}{g2}(idxs2)];
                    cpop{i}{g2} = [cpop{i}{g2} cpop{i}{g1}(idxs1)];
                    cpop{i}{g1}(idxs1) = [];
                    cpop{i}{g2}(idxs2) = [];
                    cpop{i}{g1} = sort(cpop{i}{g1});
                    cpop{i}{g2} = sort(cpop{i}{g2});                    
                end
                
                if ahelper2(i) % Did a parent mutate ?
                    ahelper(ahelper2(i)) = 0;
                    ahelper2(i) = 0;
                end
            end
        end
    end

    %% Make new population
    
    % Select only new individuals (remove candidate parents).
    pop_c(:,1) = cpop(find(~ahelper2(1:ncand-1)));
    nnew = size(pop_c,1);
    clear cpop ahelper2;

    % Compute fitness for new individuals.
    afitness_c = fitness(pop_c, asubscriptions, aclients, aclient_bin_subs, aweights);
    ngengap = 0; % Generation Gap.

    if useElitism > 0
        % Select individuals from parents and candidates into new population.   
        [afitness_p, ndx_p] = sort([afitness_c afitness_p(1:useElitism) afitness_p(useElitism + 1 : npopsize - nnew)]);
        afitness_p = afitness_p(1:npopsize);

        for i = 1:npopsize
            if ndx_p(i) > nnew
                npop(i,1) = pop(ndx_p(i) - nnew);
            else
                npop(i,1) = pop_c(ndx_p(i));
                ngengap = ngengap + 1;
            end
        end
    else
        % Select candidate parents.
        acandp = find(ahelper);
        
        if npopsize > nnew + length(acandp) % Do we need to complete population ?
            afillp = find(~ahelper);
            acandp = [acandp afillp(1:npopsize - nnew - length(acandp))];
            clear afillp;
        end

        [afitness_p, ndx_p] = sort([afitness_c afitness_p(acandp)]);

        for i = 1:npopsize
            if ndx_p(i) > nnew
                npop(i,1) = pop(acandp(ndx_p(i) - nnew));
            else
                npop(i,1) = pop_c(ndx_p(i));
                ngengap = ngengap + 1;                
            end
        end

        clear acandp;
    end

    clear pop pop_c ndx_p ahelper;
    pop = npop;
    clear npop;

    nIteration = nIteration + 1;
    
    if afitness_p(1) < cost
        iter = nIteration;
        cost = afitness_p(1);
        solution = pop{1};
    end
    
    agendata(nIteration,:) = [cost std(afitness_p,1) ngengap etime(clock,stime)];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Selection functions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Tournament selection.

function acouples = tournament(afitness_p, ncrossovers, k)
global npopsize
acouples = uint8(zeros(2, ncrossovers));
ahelper = uint8(zeros(npopsize, npopsize));
    
for i = 1:ncrossovers
    % 1st parent
    p1 = challenge(k);
    acouples(1,i) = p1;
    
    % 2nd parent
    while (1)
        p2 = challenge(k);
        
        if ~(p1 == p2 | ahelper(p1,p2) | ahelper(p2,p1))
            ahelper(p1,p2) = 1;
            ahelper(p2,p1) = 1;
            acouples(2,i) = p2;
            break;
        end        
    end
end

function ind = challenge(k)
global npopsize
p1 = fix(rand .* npopsize) + 1;
p2 = fix(rand .* npopsize) + 1;
    
if rand < k
    if p1 > p2
        ind = p2;
    else
        ind = p1;
    end
else
    if p1 > p2
        ind = p1;
    else
        ind = p2;
    end        
end

% Wheel selection.

function acouples = wheel(afitness_p, ncrossovers)
global npopsize
aRW = cumsum(1 ./ afitness_p); % Since fitness() computes cost we must use 1/cost.
nTF = aRW(end);
acouples = uint8(zeros(2, ncrossovers));
ahelper = uint8(zeros(npopsize, npopsize));
    
for i = 1:ncrossovers
    % 1st parent
    tmp = find((rand .* nTF) <= aRW);
    p1 = tmp(1);
    acouples(1,i) = p1;
        
    % 2nd parent
    while (1)
        tmp = find((rand .* nTF) <= aRW);
        p2 = tmp(1);
            
        if ~(p1 == p2 | ahelper(p1,p2) | ahelper(p2,p1))
            ahelper(p1,p2) = 1;
            ahelper(p2,p1) = 1;
            acouples(2,i) = p2;
            break;
        end
    end
end


function idxs = mutate_at(s_idx, e_idx)
d = e_idx - s_idx + 1;
mask = fix(rand(1,d) + 0.5);
idxs = s_idx:e_idx;
idxs = idxs(find(mask));